#------------------------------------------------------------------------------
# Copyright (c) 2013-2019, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
from types import CodeType

import bytecode as bc


class CodeTracer(object):
    """ A base class for implementing code tracers.

    This class defines the interface for a code tracer object, which is
    an object which can be passed as the first argument to a code object
    which has been transformed to enable tracing. Methods on the tracer
    are called with relevant arguments from the Python stack when that
    particular code segment is executing. The return value of a tracer
    method is ignored; exceptions are propagated.

    """
    __slots__ = ()

    def dynamic_load(self, obj, attr, value):
        """ Called when an dynamic attribute is loaded.

        This method is called by the dynamic scope when it loads an
        attribute from some object in the object hierarchy due to the
        execution of the LOAD_NAME opcode.

        Parameters
        ----------
        obj : Object
            The Enaml object which owns the dynamically scoped attr.

        attr : str
            The name of the attribute which was loaded.

        value : object
            The value which was loaded.

        """
        pass

    def load_attr(self, obj, attr):
        """ Called before the LOAD_ATTR opcode is executed.

        Parameters
        ----------
        obj : object
            The object which owns the attribute.

        attr : string
            The attribute being loaded.

        """
        pass

    def call_function(self, func, argtuple, argspec):
        """ Called before the CALL_FUNCTION opcode is executed.

        The CALL_FUNCTION is only used for function call with only positional
        arguments in Python 3.6+ so argspec is actually just the argument
        number but is kept for backward compatibility.

        Parameters
        ----------
        func : object
            The object being called.

        argtuple : tuple
            The argument tuple from the stack.

        argspec : int
            Len of the argtuple kept for backward compatibility.

        """
        pass

    def binary_subscr(self, obj, idx):
        """ Called before the BINARY_SUBSCR opcode is executed.

        Parameters
        ----------
        obj : object
            The object being indexed.

        idx : object
            The index.

        """
        pass

    def get_iter(self, obj):
        """ Called before the GET_ITER opcode is executed.

        Parameters
        ----------
        obj : object
            The object which should return an iterator.

        """
        pass

    def return_value(self, value):
        """ Called before the RETURN_VALUE opcode is executed.

        Parameters
        ----------
        value : object
            The value that will be returned from the code object.

        """
        pass


class CodeInverter(object):
    """ A base class for implementing code inverters.

    This class defines the interface for a code inverter object, which is
    an object which can be passed as the first argument to a code object
    which has been transformed to enable inversion. The methods on the
    inverter are called with relevant arguments from the Python stack
    when that particular code segment is executing. The return values of
    a tracer method is ignored; exceptions are propagated.

    The default behavior of an inverter is to raise. Implementations
    must provide their own code in order to enable inversion.

    """
    __slots__ = ()

    def fail(self):
        """ Called by handlers to raise an inversion exception.

        """
        raise RuntimeError("can't assign to expression")

    def load_name(self, name, value):
        """ Called before the LOAD_NAME opcode is executed.

        This method should perform a STORE_NAME operation.

        Parameters
        ----------
        name : string
            The name being loaded.

        value : object
            The value to store.

        """
        self.fail()

    def load_attr(self, obj, attr, value):
        """ Called before the LOAD_ATTR opcode is executed.

        This method should perform a STORE_ATTR operation.

        Parameters
        ----------
        obj : object
            The object which owns the attribute.

        attr : string
            The attribute being loaded.

        value : object
            The value to store

        """
        self.fail()

    def call_function(self, func, argtuple, argspec, value):
        """ Called before the CALL_FUNCTION opcode is executed.

        The CALL_FUNCTION is only used for function call with only positional
        arguments in Python 3.6+ so argspec is actually just the argument
        number but is kept for backward compatibility.

        This method should perform an appropriate store operation.

        Parameters
        ----------
        func : object
            The object being called.

        argtuple : tuple
            The argument tuple from the stack.

        argspec : int
            Len of the argtuple kept for backward compatibility.

        value : object
            The value to store.

        Notes
        -----
        The semantics of the arguments is identical to the method
        `call_function` on the `CodeTracer` type.

        """
        self.fail()

    def binary_subscr(self, obj, idx, value):
        """ Called before the BINARY_SUBSCR opcode is executed.

        This method should perform a STORE_SUBSCR operation.

        Parameters
        ----------
        obj : object
            The object being indexed.

        idx : object
            The index.

        value : object
            The value to store.

        """
        self.fail()


def inject_tracing(bytecode, nested=False):
    """ Inject tracing code into the given code list.

    This will inject the bytecode operations required to trace the
    execution of the code using a `CodeTracer` object. The generated
    opcodes expect a fast local '_[tracer]' to be available when the
    code is executed.

    Parameters
    ----------
    bytecode : bytecode.Bytecode
        The bytecode to modify.
    nested : bool
        Is the code modified defined inside an already traced code in which
        case the tracer is in the local namespace and not the fast locals.

    Returns
    -------
    result : list
        A *new* list of code ops which implement the desired behavior.

    """
    # If the code is nested into another already traced code, we need to use
    # LOAD_NAME to access the tracer rather than LOAD_FAST
    tracer_op = "LOAD_NAME" if nested else "LOAD_FAST"

    # This builds a mapping of code idx to a list of ops, which are the
    # tracing bytecode instructions which will be inserted into the code
    # object being transformed. The ops assume that a tracer object is
    # available in the fast locals using a non-clashable name. All of
    # the ops have a net-zero effect on the execution stack. Provided
    # that the tracer has no visible side effects, the tracing is
    # transparent.
    inserts = {}
    for idx, instr in enumerate(bytecode):
        # Filter out SetLineno and Label
        if not isinstance(instr, bc.Instr):
            continue
        i_name = instr.name
        i_arg = instr.arg
        if i_name == "LOAD_ATTR":
            tracing_code = [                           # obj
                bc.Instr("DUP_TOP"),                   # obj -> obj
                bc.Instr(tracer_op, '_[tracer]'),      # obj -> obj -> tracer
                bc.Instr("LOAD_ATTR", 'load_attr'),    # obj -> obj -> tracefunc
                bc.Instr("ROT_TWO"),                   # obj -> tracefunc -> obj
                bc.Instr("LOAD_CONST", i_arg),         # obj -> tracefunc -> obj -> attr
                bc.Instr("CALL_FUNCTION", 0x0002),     # obj -> retval
                bc.Instr("POP_TOP"),                   # obj
            ]
            inserts[idx] = tracing_code
        elif i_name == "CALL_FUNCTION":
            # From Python 3.6, CALL_FUNCTION is only used for positional
            # arguments and the argument is directly the number of arguments
            # on the stack.
            n_stack_args = i_arg
            tracing_code = [                                         # func -> arg(0) -> arg(1) -> ... -> arg(n-1)
                bc.Instr("BUILD_TUPLE", n_stack_args),       # func -> argtuple
                bc.Instr("DUP_TOP_TWO"),                     # func -> argtuple -> func -> argtuple
                bc.Instr(tracer_op, '_[tracer]'),            # func -> argtuple -> func -> argtuple -> tracer
                bc.Instr("LOAD_ATTR", 'call_function'),      # func -> argtuple -> func -> argtuple -> tracefunc
                bc.Instr("ROT_THREE"),                       # func -> argtuple -> tracefunc -> func -> argtuple
                bc.Instr("LOAD_CONST", i_arg),               # func -> argtuple -> tracefunc -> func -> argtuple -> argspec
                bc.Instr("CALL_FUNCTION", 0x0003),           # func -> argtuple -> retval
                bc.Instr("POP_TOP"),                         # func -> argtuple
                bc.Instr("UNPACK_SEQUENCE", n_stack_args),   # func -> arg(n-1) -> arg(n-2) -> ... -> arg(0)
                bc.Instr("BUILD_TUPLE", n_stack_args),       # func -> reversedargtuple
                bc.Instr("UNPACK_SEQUENCE", n_stack_args),   # func -> arg(0) -> arg(1) -> ... -> arg(n-1)
            ]
            inserts[idx] = tracing_code

        # We do not trace CALL_FUNCTION_KW and CALL_FUNCTION_EX since
        # tracing and inverting only require to detect getattr and setattr
        # both in this project and in traits-enaml. Those two are always called
        # using the CALL_FUNCTION bytecode instruction.

        elif i_name == "BINARY_SUBSCR":
            tracing_code = [                                     # obj -> idx
                bc.Instr("DUP_TOP_TWO"),                 # obj -> idx -> obj -> idx
                bc.Instr(tracer_op, '_[tracer]'),        # obj -> idx -> obj -> idx -> tracer
                bc.Instr("LOAD_ATTR", 'binary_subscr'),  # obj -> idx -> obj -> idx -> tracefunc
                bc.Instr("ROT_THREE"),                   # obj -> idx -> tracefunc -> obj -> idx
                bc.Instr("CALL_FUNCTION", 0x0002),       # obj -> idx -> retval
                bc.Instr("POP_TOP"),                     # obj -> idx
            ]
            inserts[idx] = tracing_code
        elif i_name == "GET_ITER":
            tracing_code = [                               # obj
                bc.Instr("DUP_TOP"),               # obj -> obj
                bc.Instr(tracer_op, '_[tracer]'),  # obj -> obj -> tracer
                bc.Instr("LOAD_ATTR", 'get_iter'), # obj -> obj -> tracefunc
                bc.Instr("ROT_TWO"),               # obj -> tracefunc -> obj
                bc.Instr("CALL_FUNCTION", 0x0001), # obj -> retval
                bc.Instr("POP_TOP"),               # obj
            ]
            inserts[idx] = tracing_code
        elif i_name == "RETURN_VALUE":
            tracing_code = [
                bc.Instr("DUP_TOP"),                   # obj
                bc.Instr(tracer_op, '_[tracer]'),      # obj -> obj -> tracer
                bc.Instr("LOAD_ATTR", 'return_value'), # obj -> obj -> tracefunc
                bc.Instr("ROT_TWO"),                   # obj -> tracefunc -> obj
                bc.Instr("CALL_FUNCTION", 0x0001),     # obj -> retval
                bc.Instr("POP_TOP"),                   # obj
            ]
            inserts[idx] = tracing_code
        elif isinstance(i_arg, CodeType):
            # Inject tracing in nested code object if they use their parent
            # locals.
            inner = bc.Bytecode.from_code(i_arg)
            if not inner.flags & bc.CompilerFlags.NEWLOCALS:
                instr.arg = inject_tracing(inner, nested=True).to_code()

    # Create a new code list which interleaves the generated code with
    # the original code at the appropriate location.
    new_code = bytecode.copy()
    new_code.clear()
    for idx, code_op in enumerate(bytecode):
        if idx in inserts:
            new_code.extend(inserts[idx])
        new_code.append(code_op)

    return new_code


def inject_inversion(bytecode):
    """ Inject inversion code into the given bytecode.

    This will inject the bytecode operations required to invert the
    execution of the code using a `CodeInverter` object. The generated
    opcodes expect the fast local '_[inverter]' and '_[value]' to be
    available when the code is executed.

    Parameters
    ----------
    bytecode : bc.Bytecode
        The list of code ops to modify.

    Returns
    -------
    result : list
        A *new* list of code ops which implement the desired behavior.

    Raises
    ------
    ValueError
        The given code is not suitable for inversion.

    """
    instr = bytecode[-2]
    i_name, i_arg = instr.name, instr.arg
    new_code = bytecode[:-2]
    if i_name == "LOAD_NAME" and len(bytecode) == 2:
        new_code.extend([                             #
            bc.Instr("LOAD_FAST", '_[inverter]'),     # inverter
            bc.Instr("LOAD_ATTR", 'load_name'),       # invertfunc
            bc.Instr("LOAD_CONST", i_arg),            # invertfunc -> name
            bc.Instr("LOAD_FAST", '_[value]'),        # invertfunc -> name - > value
            bc.Instr("CALL_FUNCTION", 0x0002),        # retval
           bc.Instr ("RETURN_VALUE"),                 #
        ])
    elif i_name == "LOAD_ATTR":
        new_code.extend([                   # obj
            bc.Instr("LOAD_FAST", '_[inverter]'),     # obj -> inverter
            bc.Instr("LOAD_ATTR", 'load_attr'),       # obj -> invertfunc
            bc.Instr("ROT_TWO"),                      # invertfunc -> obj
            bc.Instr("LOAD_CONST", i_arg),            # invertfunc -> obj -> attr
            bc.Instr("LOAD_FAST", '_[value]'),        # invertfunc -> obj -> attr -> value
            bc.Instr("CALL_FUNCTION", 0x0003),        # retval
            bc.Instr("RETURN_VALUE"),                 #
        ])

    # In Python 3.6+ CALL_FUNCTION is only used for calls with positional arguments
    # and the argument of the opcode is teh number of argument on the stack.
    elif i_name == "CALL_FUNCTION":
        n_stack_args = i_arg
        new_code.extend([                             # func -> arg(0) -> arg(1) -> ... -> arg(n-1)
            bc.Instr("BUILD_TUPLE", n_stack_args),    # func -> argtuple
            bc.Instr("LOAD_FAST", '_[inverter]'),     # func -> argtuple -> inverter
            bc.Instr("LOAD_ATTR", 'call_function'),   # func -> argtuple -> invertfunc
            bc.Instr("ROT_THREE"),                    # invertfunc -> func -> argtuple
            bc.Instr("LOAD_CONST", i_arg),            # invertfunc -> func -> argtuple -> argspec
            bc.Instr("LOAD_FAST", '_[value]'),        # invertfunc -> func -> argtuple -> argspec -> value
            bc.Instr("CALL_FUNCTION", 0x0004),        # retval
            bc.Instr("RETURN_VALUE"),                 #
        ])

    # We do not trace CALL_FUNCTION_KW and CALL_FUNCTION_EX since tracing and
    # inverting only require to detect getattr and setattr both in this project
    # and in traits-enaml. Those two are always called using the CALL_FUNCTION
    # bytecode instruction.

    elif i_name == "BINARY_SUBSCR":
        new_code.extend([                             # obj -> index
            bc.Instr("LOAD_FAST", '_[inverter]'),     # obj -> index -> inverter
            bc.Instr("LOAD_ATTR", 'binary_subscr'),   # obj -> index -> invertfunc
            bc.Instr("ROT_THREE"),                    # invertfunc -> obj -> index
            bc.Instr("LOAD_FAST", '_[value]'),        # invertfunc -> obj -> index -> value
            bc.Instr("CALL_FUNCTION", 0x0003),        # retval
            bc.Instr("RETURN_VALUE"),                 #
        ])
    else:
        raise ValueError("can't invert code")

    return new_code



#------------------------------------------------------------------------------
# Copyright (c) 2019, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
from atom.api import Bool, ForwardTyped, Typed, observe

from enaml.core.declarative import d_

from .toolkit_object import ToolkitObject, ProxyToolkitObject


class ProxyButtonGroup(ProxyToolkitObject):
    """ The abstract defintion of a proxy ButtonGroup object.

    """
    #: A reference to the ButtonGroup declaration.
    declaration = ForwardTyped(lambda: ButtonGroup)

    def set_exclusive(self, exclusive):
        raise NotImplementedError

    def add_button(self, button):
        raise NotImplementedError

    def remove_button(self, button):
        raise NotImplementedError


class ButtonGroup(ToolkitObject):
    """ A way to declare a group of buttons.

    This allows to group buttons even though they belong to different
    container. Note that if a button belongs to a ButtonGroup the rules
    about buttons sharing a container being in the same group do not apply.d2

    """
    #: Set of members belonging to the group.
    #: This value should be considered read-only for users.
    group_members = Typed(set, ())

    #: Can only a single button in the group be checked at a time.
    exclusive = d_(Bool(True))

    #: A reference to the ProxyButtonGroup object.
    proxy = Typed(ProxyButtonGroup)

    @observe("exclusive")
    def _update_proxy(self, change):
        super(ButtonGroup, self)._update_proxy(change)




#------------------------------------------------------------------------------
# Copyright (c) 2013, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
from .action_item import ActionItem
from .autostart import Autostart
from .branding import Branding
from .item_group import ItemGroup
from .menu_item import MenuItem
from .ui_workbench import UIWorkbench
from .workspace import Workspace



#------------------------------------------------------------------------------
# Copyright (c) 2013, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
from atom.api import Atom, Enum, Bytes, Value, Coerced

from enaml.layout.geometry import Size


class Image(Atom):
    """ An object representing an image.

    Once an image is created it should be treated as read only. User
    code should create a new image object if the parameters need to
    be changed.

    """
    #: The format of the image. By default, the consumer of the image
    #: will probe the header to automatically infer a type.
    #: Base64 images need to be decoded using the base64.b64decode function
    #: from the standard library.
    format = Enum(
        'auto',     # Automatically determine the image format
        'png',      # Portable Network Graphics
        'jpg',      # Joint Photographic Experts Group
        'gif',      # Graphics Interchange Format
        'bmp',      # Windows Bitmap
        'xpm',      # X11 Pixmap
        'xbm',      # X11 Bitmap
        'pbm',      # Portable Bitmap
        'pgm',      # Portable Graymap
        'ppm',      # Portable Pixmap
        'tiff',     # Tagged Image File Format
        'argb32',   # Raw data in the 0xAARRGGBB format.
                    # The `raw_size` of the image must be provided.
    )

    #: The (width, height) raw size of the image. This must be provided
    #: for images where the size is not encoded in the data stream.
    raw_size = Coerced(Size, (0, 0))

    #: The (width, height) size of the image. An invalid size indicates
    #: that the size of the image should be automatically inferred. A
    #: valid size indicates that the toolkit image should be scaled to
    #: the specified size.
    size = Coerced(Size, (-1, -1))

    #: The aspect ratio mode to use when the toolkit scales the image.
    aspect_ratio_mode = Enum('ignore', 'keep', 'keep_by_expanding')

    #: The transform mode to use when the toolkit scales the image.
    transform_mode = Enum('smooth', 'fast')

    # XXX this needs to be augmented to support arrays.
    #: The bytestring holding the data for the image.
    data = Bytes()

    #: Storage space for use by a toolkit backend to use as needed.
    #: This should not typically be manipulated by user code.
    _tkdata = Value()



#------------------------------------------------------------------------------
# Copyright (c) 2013, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
import re

from enaml.styling import StyleCache


_grad_re = re.compile(r'(lineargradient|radialgradient)')

_alert_re = re.compile(r'alert\((-?[_a-zA-Z][_a-zA-Z0-9-]*)\)')


def _translate_gradient(v):
    return _grad_re.sub(r'q\1', v)


_KNOWN_FIELDS = set([
    'alternate-background-color',
    'background',
    'background-clip',
    'background-color',
    'border',
    'border-top',
    'border-right',
    'border-bottom',
    'border-left',
    'border-color',
    'border-top-color',
    'border-right-color',
    'border-bottom-color',
    'border-left-color',
    'border-radius',
    'border-top-left-radius',
    'border-top-right-radius',
    'border-bottom-right-radius',
    'border-bottom-left-radius',
    'border-style',
    'border-top-style',
    'border-right-style',
    'border-bottom-style',
    'border-left-style',
    'border-width',
    'border-top-width',
    'border-right-width',
    'border-bottom-width',
    'border-left-width',
    'bottom',
    'color',
    'font',
    'font-family',
    'font-size',
    'font-style',
    'font-weight',
    'height',
    'icon-size',
    'left',
    'line-edit-password-character',
    'margin',
    'margin-top',
    'margin-right',
    'margin-bottom',
    'margin-left',
    'max-height',
    'max-width',
    'min-height',
    'min-width',
    'padding',
    'padding-top',
    'padding-right',
    'padding-bottom',
    'padding-left',
    'position',
    'right',
    'selection-background-color',
    'selection-color',
    'spacing',
    'subcontrol-origin',
    'subcontrol-position',
    'text-align',
    'text-decoration',
    'top',
    'width',
])


_MAY_HAVE_GRADIENT = set([
    'background',
    'background-color',
    'border',
    'border-top',
    'border-right',
    'border-bottom',
    'border-left',
    'border-color',
    'border-top-color',
    'border-right-color',
    'border-bottom-color',
    'border-left-color',
    'color',
    'selection-background-color',
    'selection-color',
])


def _translate_setter(setter):
    field = setter.field
    if field in _KNOWN_FIELDS:
        value = setter.value
        if value:
            if field in _MAY_HAVE_GRADIENT:
                value = _translate_gradient(value)
            value = '    %s: %s;' % (field, value)
            return value


def _translate_style_body(style):
    translated = []
    for setter in style.setters():
        tks = StyleCache.toolkit_setter(setter, _translate_setter)
        if tks is not None:
            translated.append(tks)
    return '\n'.join(translated)


def translate_style(name, style):
    parts = ['#%s' % name]
    if style.pseudo_element:
        root = parts.pop()
        for pe in style.pseudo_element.split(','):
            parts.append(root + '::%s' % pe.strip())
    if style.pseudo_class:
        these = parts[:]
        parts = []
        for pc in style.pseudo_class.split(','):
            pc = pc.strip()
            for this in these:
                parts.append(this + ':%s' % pc)
    selector = ','.join(parts)
    body = '{\n%s\n}' % _translate_style_body(style)
    return '%s %s' % (selector, body)


#------------------------------------------------------------------------------
# Dock Area Styling
#------------------------------------------------------------------------------
def _basic_pc(root, pc):
    if pc:
        root += ':%s' % pc
    return root


def _base_dock_area(name, pc):
    rest = []
    floating = False
    for part in pc.split(':'):
        if part == 'floating':
            floating = True
        else:
            rest.append(part)
    if floating:
        root = 'QDockWindow > QDockArea'
    else:
        root = 'QDockArea'
    return _basic_pc(root, ':'.join(rest))


def _dock_bar(name, pc):
    return _basic_pc('QDockBar', pc)


def _dock_bar_button(name, pc):
    return _maybe_alert('QDockBarButton', name, pc)


_position_map = {'top': '0', 'right': '1', 'bottom': '2', 'left': '3'}
def _dock_bar_handle(name, pc):
    rest = []
    position = None
    for part in pc.split(':'):
        if part in _position_map:
            position = _position_map[part]
        else:
            rest.append(part)
    if position is not None:
        root = 'QDockBarItem[position="%s"] QDockBarItemHandle' % position
    else:
        root = 'QDockBarItem QDockBarItemHandle'
    return _basic_pc(root, ':'.join(rest))


def _dock_container(name, pc):
    rest = []
    tabbed = False
    floating = False
    for part in pc.split(':'):
        if part == 'tabbed':
            tabbed = True
        elif part == 'floating':
            floating = True
        else:
            rest.append(part)
    if tabbed:
        root = 'QDockTabWidget QDockContainer'
    elif floating:
        root = 'QDockContainer[floating="true"]'
    else:
        root = 'QDockContainer'
    return _maybe_alert(root, name, ':'.join(rest))


def _dock_window(name, pc):
    return _basic_pc('QDockWindow', pc)


def _dock_window_button(name, pc):
    return _basic_pc('QDockWindowButtons QBitmapButton', pc)


def _dock_window_close_button(name, pc):
    return _basic_pc('QBitmapButton#dockwindow-close-button', pc)


def _dock_window_link_button(name, pc):
    return _basic_pc('QBitmapButton#dockwindow-link-button', pc)


def _dock_window_maximize_button(name, pc):
    return _basic_pc('QBitmapButton#dockwindow-maximize-button', pc)


def _dock_window_restore_button(name, pc):
    return _basic_pc('QBitmapButton#dockwindow-restore-button', pc)


def _rubber_band(name, pc):
    return _basic_pc('QDockRubberBand', pc)


def _splitter_handle(name, pc):
    return _basic_pc('QDockSplitter::handle', pc)


def _tab_bar(name, pc):
    return _basic_pc('QDockTabBar', pc)


def _tab_bar_tab(name, pc):
    rest = []
    alert = ''
    for part in pc.split(':'):
        match = _alert_re.match(part)
        if match is not None:
            alert = match.group(1)
        else:
            rest.append(part)
    if alert:
        root = 'QDockTabBar[alert="%s"]::tab' % alert
    else:
        root = 'QDockTabBar::tab'
    return _basic_pc(root, ':'.join(rest))


def _tab_bar_close_button(name, pc):
    return _basic_pc('QBitmapButton#docktab-close-button', pc)


def _tab_bar_corner(name, pc):
    return _basic_pc('QDockTabWidget::right-corner', pc)


def _tab_bar_corner_bar(name, pc):
    return _basic_pc('QDockTitleBar#docktab-corner-widget', pc)


def _tab_bar_corner_button(name, pc):
    return _basic_pc('QDockTitleBar#docktab-corner-widget > QBitmapButton', pc)


def _maybe_alert(root, name, pc):
    rest = []
    alert = ''
    for part in pc.split(':'):
        match = _alert_re.match(part)
        if match is not None:
            alert = match.group(1)
        else:
            rest.append(part)
    if alert:
        root = '%s[alert="%s"]' % (root, alert)
    return _basic_pc(root, ':'.join(rest))


def _base_dock_item(name, pc):
    rest = []
    tabbed = False
    floating = False
    for part in pc.split(':'):
        if part == 'tabbed':
            tabbed = True
        elif part == 'floating':
            floating = True
        else:
            rest.append(part)
    if tabbed:
        root = 'QDockTabWidget QDockItem'
    elif floating:
        root = 'QDockContainer[floating="true"] QDockItem'
    else:
        root = 'QDockItem'
    return _maybe_alert(root, name, ':'.join(rest))


def _title_bar(name, pc):
    return _maybe_alert('QDockTitleBar', name, pc)


def _title_bar_label(name, pc):
    return _maybe_alert('QDockTitleBar > QTextLabel', name, pc)


def _title_bar_button(name, pc):
    return _basic_pc('QDockTitleBar > QBitmapButton', pc)


def _title_bar_close_button(name, pc):
    return _basic_pc('QBitmapButton#docktitlebar-close-button', pc)


def _title_bar_link_button(name, pc):
    return _basic_pc('QBitmapButton#docktitlebar-link-button', pc)


def _title_bar_maximize_button(name, pc):
    return _basic_pc('QBitmapButton#docktitlebar-maximize-button', pc)


def _title_bar_pin_button(name, pc):
    return _basic_pc('QBitmapButton#docktitlebar-pin-button', pc)


def _title_bar_restore_button(name, pc):
    return _basic_pc('QBitmapButton#docktitlebar-restore-button', pc)


_DOCK_AREA_PSEUDO_ELEMENTS = {
    '': _base_dock_area,
    'dock-bar': _dock_bar,
    'dock-bar-button': _dock_bar_button,
    'dock-bar-handle': _dock_bar_handle,
    'dock-container': _dock_container,
    'dock-window': _dock_window,
    'dock-window-button': _dock_window_button,
    'dock-window-close-button': _dock_window_close_button,
    'dock-window-link-button': _dock_window_link_button,
    'dock-window-maximize-button': _dock_window_maximize_button,
    'dock-window-restore-button': _dock_window_restore_button,
    'rubber-band': _rubber_band,
    'splitter-handle': _splitter_handle,
    'tab-bar': _tab_bar,
    'tab-bar-tab': _tab_bar_tab,
    'tab-bar-corner': _tab_bar_corner,
    'tab-bar-corner-bar': _tab_bar_corner_bar,
    'tab-bar-corner-button': _tab_bar_corner_button,
    'tab-bar-close-button': _tab_bar_close_button,
}


_DOCK_ITEM_PSEUDO_ELEMENTS = {
    '': _base_dock_item,
    'title-bar': _title_bar,
    'title-bar-label': _title_bar_label,
    'title-bar-button': _title_bar_button,
    'title-bar-close-button': _title_bar_close_button,
    'title-bar-link-button': _title_bar_link_button,
    'title-bar-maximize-button': _title_bar_maximize_button,
    'title-bar-pin-button': _title_bar_pin_button,
    'title-bar-restore-button': _title_bar_restore_button,
}


def _dock_style_selector(name, style, elements):
    handlers = []
    for pe in style.pseudo_element.split(','):
        handler = elements.get(pe.strip())
        if handler:
            handlers.append(handler)
    if not handlers:
        return
    parts = []
    for pc in style.pseudo_class.split(','):
        pc = pc.strip()
        for handler in handlers:
            parts.append(handler(name, pc))
    return ', '.join(parts)


def translate_dock_area_style(name, style):
    selector = _dock_style_selector(name, style, _DOCK_AREA_PSEUDO_ELEMENTS)
    if not selector:
        return
    body = '{\n%s\n}' % _translate_style_body(style)
    return '%s %s' % (selector, body)


def translate_dock_item_style(name, style):
    selector = _dock_style_selector(name, style, _DOCK_ITEM_PSEUDO_ELEMENTS)
    if not selector:
        return
    body = '{\n%s\n}' % _translate_style_body(style)
    return '%s %s' % (selector, body)



#------------------------------------------------------------------------------
# Copyright (c) 2013, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
from atom.api import Bool, Dict, Typed, Str

from enaml.core.declarative import Declarative, d_
from enaml.icon import Icon


class ActionItem(Declarative):
    """ A declarative class for defining a workbench action item.

    """
    #: The "/" separated path to this item in the menu bar.
    path = d_(Str())

    #: The parent menu group to which this action item belongs.
    group = d_(Str())

    #: The action item will appear before this item in its group.
    before = d_(Str())

    #: The action item will appear after this item in its group.
    after = d_(Str())

    #: The id of the Command invoked by the action.
    command = d_(Str())

    #: The user parameters to pass to the command handler.
    parameters = d_(Dict())

    #: The display label for the action.
    label = d_(Str())

    #: The shortcut keybinding for the action. e.g. Ctrl+C
    shortcut = d_(Str())

    #: Whether or not the action is visible.
    visible = d_(Bool(True))

    #: Whether or not the action is enabled.
    enabled = d_(Bool(True))

    #: Whether or not the action is checkable.
    checkable = d_(Bool(False))

    #: Whether or not the checkable action is checked.
    checked = d_(Bool(False))

    #: The default display icon for the action.
    icon = d_(Typed(Icon))

    #: The tooltip for the action.
    tool_tip = d_(Str())

    #: The statustip for the action.
    status_tip = d_(Str())



#------------------------------------------------------------------------------
# Copyright (c) 2013-2018, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------

#: The global dict of registered stylesheets. This dict should never be
#: modified directly. Use the functions 'register_style_sheet' and
#: 'get_style_sheet' instead.
_style_sheets = {}


def get_style_sheet(name):
    """ This is a deprecated function.

    """
    return _style_sheets.get(name, u'')


def register_style_sheet(name, sheet):
    """ This is a deprecated function.

    """
    import warnings
    msg = "Toolkit specific styling for the DockArea is deprecated "
    msg += "and will be removed in Enaml version 1.0. Use stylesheets "
    msg += "to style the Dock Area instead."
    warnings.warn(msg, FutureWarning, stacklevel=2)

    assert isinstance(sheet, str), 'style sheet must be a string'
    if name in _style_sheets:
        raise ValueError("'%s' style is already registered" % name)
    _style_sheets[name] = sheet



#------------------------------------------------------------------------------
# Copyright (c) 2013, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
from __future__ import unicode_literals

from atom.api import Typed

from enaml.application import Application
from enaml.workbench.extension import Extension
from enaml.workbench.plugin import Plugin

from .action_item import ActionItem
from .autostart import Autostart
from .branding import Branding
from .menu_helper import create_menus
from .menu_item import MenuItem
from .window_model import WindowModel
from .workspace import Workspace

import enaml
with enaml.imports():
    from .workbench_window import WorkbenchWindow


ACTIONS_POINT = u'enaml.workbench.ui.actions'

APPLICATION_FACTORY_POINT = u'enaml.workbench.ui.application_factory'

BRANDING_POINT = u'enaml.workbench.ui.branding'

WINDOW_FACTORY_POINT = u'enaml.workbench.ui.window_factory'

WORKSPACES_POINT = u'enaml.workbench.ui.workspaces'

AUTOSTART_POINT = u'enaml.workbench.ui.autostart'


class UIPlugin(Plugin):
    """ The main UI plugin class for the Enaml studio.

    The ui plugin manages the extension points for user contributions
    to the main window.

    """
    def start(self):
        """ Start the plugin life-cycle.

        This method is called by the framework at the appropriate time.
        It should never be called by user code.

        """
        self._create_application()
        self._create_model()
        self._create_window()
        self._bind_observers()
        self._start_autostarts()

    def stop(self):
        """ Stop the plugin life-cycle.

        This method is called by the framework at the appropriate time.
        It should never be called by user code.

        """
        self._stop_autostarts()
        self._unbind_observers()
        self._destroy_window()
        self._release_model()
        self._release_application()

    @property
    def window(self):
        """ Get a reference to the primary window.

        """
        return self._window

    @property
    def workspace(self):
        """ Get a reference to the currently active workspace.

        """
        return self._model.workspace

    def show_window(self):
        """ Ensure the underlying window object is shown.

        """
        self._window.show()

    def hide_window(self):
        """ Ensure the underlying window object is hidden.

        """
        self._window.hide()

    def start_application(self):
        """ Start the application event loop.

        """
        self._application.start()

    def stop_application(self):
        """ Stop the application event loop.

        """
        self._application.stop()

    def close_window(self):
        """ Close the underlying workbench window.

        """
        self._window.close()

    def close_workspace(self):
        """ Close and dispose of the currently active workspace.

        """
        self._workspace_extension = None
        self._model.workspace.stop()
        self._model.workspace.workbench = None
        self._model.workspace = Workspace()

    def select_workspace(self, extension_id):
        """ Select and start the workspace for the given extension id.

        The current workspace will be stopped and released.

        """
        target = None
        workbench = self.workbench
        point = workbench.get_extension_point(WORKSPACES_POINT)
        for extension in point.extensions:
            if extension.qualified_id == extension_id:
                target = extension
                break

        if target is None:
            msg = "'%s' is not a registered workspace extension"
            raise ValueError(msg % extension_id)

        if target is self._workspace_extension:
            return

        old_workspace = self._model.workspace
        old_workspace.stop()
        old_workspace.workbench = None

        self._workspace_extension = target
        new_workspace = self._create_workspace(target)

        new_workspace.workbench = workbench
        new_workspace.start()
        self._model.workspace = new_workspace

    #--------------------------------------------------------------------------
    # Private API
    #--------------------------------------------------------------------------
    #: The application provided by an ApplicationFactory extension.
    _application = Typed(Application)

    #: The window object provided by a WindowFactory extension.
    _window = Typed(WorkbenchWindow)

    #: The view model object used to drive the window.
    _model = Typed(WindowModel)

    #: The currently activate branding extension object.
    _branding_extension = Typed(Extension)

    #: The currently activate workspace extension object.
    _workspace_extension = Typed(Extension)

    #: The currently active action extension objects.
    _action_extensions = Typed(dict, ())

    def _create_application(self):
        """ Create the Application object for the ui.

        This will load the highest ranking extension to the application
        factory extension point, and use it to create the instance.

        If an application object already exists, that application will
        be used instead of any defined by a factory, since there can be
        only one application per-process.

        """
        if Application.instance() is not None:
            self._application = Application.instance()
            return

        workbench = self.workbench
        point = workbench.get_extension_point(APPLICATION_FACTORY_POINT)
        extensions = point.extensions
        if not extensions:
            msg = "no contributions to the '%s' extension point"
            raise RuntimeError(msg % APPLICATION_FACTORY_POINT)

        extension = extensions[-1]
        if extension.factory is None:
            msg = "extension '%s' does not declare an application factory"
            raise ValueError(msg % extension.qualified_id)

        application = extension.factory()
        if not isinstance(application, Application):
            msg = "extension '%s' created non-Application type '%s'"
            args = (extension.qualified_id, type(application).__name__)
            raise TypeError(msg % args)

        self._application = application

    def _create_model(self):
        """ Create and initialize the model which drives the window.

        """
        self._model = WindowModel()
        self._refresh_branding()
        self._refresh_actions()

    def _create_window(self):
        """ Create the WorkbenchWindow object for the workbench.

        This will load the highest ranking extension to the window
        factory extension point, and use it to create the instance.

        """
        workbench = self.workbench
        point = workbench.get_extension_point(WINDOW_FACTORY_POINT)
        extensions = point.extensions
        if not extensions:
            msg = "no contributions to the '%s' extension point"
            raise RuntimeError(msg % WINDOW_FACTORY_POINT)

        extension = extensions[-1]
        if extension.factory is None:
            msg = "extension '%s' does not declare a window factory"
            raise ValueError(msg % extension.qualified_id)

        window = extension.factory(workbench)
        if not isinstance(window, WorkbenchWindow):
            msg = "extension '%s' created non-WorkbenchWindow type '%s'"
            args = (extension.qualified_id, type(window).__name__)
            raise TypeError(msg % args)

        window.workbench = workbench
        window.window_model = self._model
        self._window = window

    def _create_workspace(self, extension):
        """ Create the Workspace object for the given extension.

        Parameters
        ----------
        extension : Extension
            The extension object of interest.

        Returns
        -------
        result : Workspace
            The workspace object for the given extension.

        """
        if extension.factory is None:
            msg = "extension '%s' does not declare a workspace factory"
            raise ValueError(msg % extension.qualified_id)

        workspace = extension.factory(self.workbench)
        if not isinstance(workspace, Workspace):
            msg = "extension '%s' created non-Workspace type '%s'"
            args = (extension.qualified_id, type(workspace).__name__)
            raise TypeError(msg % args)

        return workspace

    def _create_action_items(self, extension):
        """ Create the action items for the extension.

        """
        workbench = self.workbench
        menu_items = extension.get_children(MenuItem)
        action_items = extension.get_children(ActionItem)
        if extension.factory:
            for item in extension.factory(workbench):
                if isinstance(item, MenuItem):
                    menu_items.append(item)
                elif isinstance(item, ActionItem):
                    action_items.append(item)
                else:
                    msg = "action extension created invalid action type '%s'"
                    raise TypeError(msg % type(item).__name__)
        return menu_items, action_items

    def _destroy_window(self):
        """ Destroy and release the underlying window object.

        """
        self._window.hide()
        self._window.destroy()
        self._window = None

    def _release_model(self):
        """ Release the underlying window model object.

        """
        self._model.workspace.stop()
        self._model = None

    def _release_application(self):
        """ Stop and release the underlyling application object.

        """
        self._application.stop()
        self._application = None

    def _refresh_branding(self):
        """ Refresh the branding object for the window model.

        """
        workbench = self.workbench
        point = workbench.get_extension_point(BRANDING_POINT)
        extensions = point.extensions
        if not extensions:
            self._branding_extension = None
            self._model.branding = Branding()
            return

        extension = extensions[-1]
        if extension is self._branding_extension:
            return

        if extension.factory:
            branding = extension.factory(workbench)
            if not isinstance(branding, Branding):
                msg = "extension '%s' created non-Branding type '%s'"
                args = (extension.qualified_id, type(branding).__name__)
                raise TypeError(msg % args)
        else:
            branding = extension.get_child(Branding, reverse=True)
            branding = branding or Branding()

        self._branding_extension = extension
        self._model.branding = branding

    def _refresh_actions(self):
        """ Refresh the actions for the workbench window.

        """
        workbench = self.workbench
        point = workbench.get_extension_point(ACTIONS_POINT)
        extensions = point.extensions
        if not extensions:
            self._action_extensions.clear()
            self._model.menus = []
            return

        menu_items = []
        action_items = []
        new_extensions = {}
        old_extensions = self._action_extensions
        for extension in extensions:
            if extension in old_extensions:
                m_items, a_items = old_extensions[extension]
            else:
                m_items, a_items = self._create_action_items(extension)
            new_extensions[extension] = (m_items, a_items)
            menu_items.extend(m_items)
            action_items.extend(a_items)

        menus = create_menus(workbench, menu_items, action_items)
        self._action_extensions = new_extensions
        self._model.menus = menus

    def _get_autostarts(self):
        """ Get the autostart extension objects.

        """
        workbench = self.workbench
        point = workbench.get_extension_point(AUTOSTART_POINT)
        extensions = sorted(point.extensions, key=lambda ext: ext.rank)

        autostarts = []
        for extension in extensions:
            autostarts.extend(extension.get_children(Autostart))

        return autostarts

    def _start_autostarts(self):
        """ Start the plugins for the autostart extension point.

        """
        workbench = self.workbench
        for autostart in self._get_autostarts():
            workbench.get_plugin(autostart.plugin_id)

    def _stop_autostarts(self):
        """ Stop the plugins for the autostart extension point.

        """
        workbench = self.workbench
        for autostart in reversed(self._get_autostarts()):
            plugin = workbench.get_plugin(autostart.plugin_id)
            plugin.stop()

    def _on_branding_updated(self, change):
        """ The observer for the branding extension point.

        """
        self._refresh_branding()

    def _on_actions_updated(self, change):
        """ The observer for the actions extension point.

        """
        self._refresh_actions()

    def _bind_observers(self):
        """ Setup the observers for the plugin.

        """
        workbench = self.workbench

        point = workbench.get_extension_point(BRANDING_POINT)
        point.observe('extensions', self._on_branding_updated)

        point = workbench.get_extension_point(ACTIONS_POINT)
        point.observe('extensions', self._on_actions_updated)

    def _unbind_observers(self):
        """ Remove the observers for the plugin.

        """
        workbench = self.workbench

        point = workbench.get_extension_point(BRANDING_POINT)
        point.unobserve('extensions', self._on_branding_updated)

        point = workbench.get_extension_point(ACTIONS_POINT)
        point.unobserve('extensions', self._on_actions_updated)



#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
import ast

from .lexer3 import Python34EnamlLexer
from .parser3 import Python3EnamlParser


class Python34EnamlParser(Python3EnamlParser):
    """Enaml parser supporting Python 3.4 syntax.

    """
    parser_id = '34'

    lexer = Python34EnamlLexer

    def p_atom13(self, p):
        ''' atom : NONE '''
        p[0] = ast.NameConstant(None)

    def p_atom14(self, p):
        ''' atom : FALSE '''
        p[0] = ast.NameConstant(False)

    def p_atom15(self, p):
        ''' atom : TRUE '''
        p[0] = ast.NameConstant(True)

    def _make_args(self, args, defaults=[], vararg=None, kwonlyargs=[],
                   kw_defaults=[], kwarg=None):
        """Build an ast node for function arguments.

        """
        return ast.arguments(args=args, defaults=defaults, vararg=vararg,
                             kwonlyargs=kwonlyargs, kw_defaults=kw_defaults,
                             kwarg=kwarg)



#------------------------------------------------------------------------------
# Copyright (c) 2013, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
from atom.api import Atom, Enum, Typed, List, Value

from .image import Image


class IconImage(Atom):
    """ An object representing an image in an icon.

    Instances of this class are used to populate the 'images' list of
    an 'Icon' instance. Instances of this class should be treated as
    read only once they are created.

    """
    #: The widget mode for which this icon should apply.
    mode = Enum('normal', 'active', 'disabled', 'selected')

    #: The widget state for which this icon should apply.
    state = Enum('off', 'on')

    #: The image to use for this icon image.
    image = Typed(Image)


class Icon(Atom):
    """ An object object representing an icon.

    Once an image is created it should be treated as read only. User
    code should create a new icon object if the parameters need to
    be changed.

    """
    #: The list of icon images which compose this icon.
    images = List(IconImage)

    #: Storage space for use by a toolkit backend to use as needed.
    #: This should not typically be manipulated by user code.
    _tkdata = Value()



#------------------------------------------------------------------------------
# Copyright (c) 2013, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
from atom.api import Typed, ForwardTyped

from .container import Container
from .widget import Widget, ProxyWidget


class ProxyStackItem(ProxyWidget):
    """ The abstract definition of a proxy StackItem object.

    """
    #: A reference to the StackItem declaration.
    declaration = ForwardTyped(lambda: StackItem)


class StackItem(Widget):
    """ A widget which can be used as an item in a Stack.

    A StackItem is a widget which can be used as a child of a Stack
    widget. It can have at most a single child widget which is an
    instance of Container.

    """
    #: A reference to the ProxyStackItem object.
    proxy = Typed(ProxyStackItem)

    def stack_widget(self):
        """ Get the stack widget defined for the item.

        The stack widget is the last child Container.

        """
        for child in reversed(self.children):
            if isinstance(child, Container):
                return child



#------------------------------------------------------------------------------
# Copyright (c) 2013-2018, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
from enaml.qt.QtCore import Qt, QSize, QPoint, QMargins, Signal
from enaml.qt.QtWidgets import (
    QWidget, QFrame, QLineEdit, QHBoxLayout, QSizePolicy
)

from .q_bitmap_button import QBitmapButton, QCheckedBitmapButton
from .q_icon_widget import QIconWidget
from .q_text_label import QTextLabel
from .xbms import (
    CLOSE_BUTTON, MAXIMIZE_BUTTON, RESTORE_BUTTON, LINKED_BUTTON,
    UNLINKED_BUTTON, PIN_BUTTON, UNPIN_BUTTON, TAB_BUTTON
)


class IDockTitleBar(QWidget):
    """ An interface class for defining a title bar.

    """
    #: A signal emitted when the maximize button is clicked.
    maximizeButtonClicked = Signal(bool)

    #: A signal emitted when the restore button is clicked.
    restoreButtonClicked = Signal(bool)

    #: A signal emitted when the close button is clicked.
    closeButtonClicked = Signal(bool)

    #: A signal emitted when the link button is toggled.
    linkButtonToggled = Signal(bool)

    #: A signal emitted when the pin button is toggled.
    pinButtonToggled = Signal(bool)

    #: A signal emitted when the tabs button is clicked.
    tabsButtonClicked = Signal(bool)

    #: A signal emitted when the title is edited by the user.
    titleEdited = Signal(str)

    #: A signal emitted when the title bar is left double clicked.
    leftDoubleClicked = Signal(QPoint)

    #: A signal emitted when the title bar is right clicked.
    rightClicked = Signal(QPoint)

    #: Do not show any buttons in the title bar.
    NoButtons = 0x0

    #: Show the maximize button in the title bar.
    MaximizeButton = 0x1

    #: Show the restore button in the title bar.
    RestoreButton = 0x2

    #: Show the close button in the title bar.
    CloseButton = 0x4

    #: Show the link button in the title bar.
    LinkButton = 0x8

    #: Show the pin button in the title bar.
    PinButton = 0x10

    #: Show the tabs menu button in the title bar.
    TabsButton = 0x20

    def buttons(self):
        """ Get the buttons to show in the title bar.

        Returns
        -------
        result : int
            An or'd combination of the buttons to show.

        """
        raise NotImplementedError

    def setButtons(self, buttons):
        """ Set the buttons to show in the title bar.

        Parameters
        ----------
        buttons : int
            An or'd combination of the buttons to show.

        """
        raise NotImplementedError

    def title(self):
        """ Get the title string of the title bar.

        Returns
        -------
        result : unicode
            The unicode title string for the title bar.

        """
        raise NotImplementedError

    def setTitle(self, title):
        """ Set the title string of the title bar.

        Parameters
        ----------
        title : unicode
            The unicode string to use for the title bar.

        """
        raise NotImplementedError

    def label(self):
        """ Get the label for the title bar.

        Returns
        -------
        result : QTextLabel
            The label for the title bar.

        """
        raise NotImplementedError

    def icon(self):
        """ Get the icon for the title bar.

        Returns
        -------
        result : QIcon
            The icon set for the title bar.

        """
        raise NotImplementedError

    def setIcon(self, icon):
        """ Set the icon for the title bar.

        Parameters
        ----------
        icon : QIcon
            The icon to use for the title bar.

        """
        raise NotImplementedError

    def iconSize(self):
        """ Get the icon size for the title bar.

        Returns
        -------
        result : QSize
            The size to use for the icons in the title bar.

        """
        raise NotImplementedError

    def setIconSize(self, size):
        """ Set the icon size for the title bar.

        Parameters
        ----------
        icon : QSize
            The icon size to use for the title bar. Icons smaller than
            this size will not be scaled up.

        """
        raise NotImplementedError

    def isLinked(self):
        """ Get whether the link button is checked.

        Returns
        -------
        result : bool
            True if the link button is checked, False otherwise.

        """
        raise NotImplementedError

    def setLinked(self, linked):
        """ Set whether or not the link button is checked.

        Parameters
        ----------
        linked : bool
            True if the link button should be checked, False otherwise.

        """
        raise NotImplementedError

    def isPinned(self):
        """ Get whether the pin button is checked.

        Returns
        -------
        result : bool
            True if the pin button is checked, False otherwise.

        """
        raise NotImplementedError

    def setPinned(self, pinned, quiet=False):
        """ Set whether or not the pin button is checked.

        Parameters
        ----------
        pinned : bool
            True if the pin button should be checked, False otherwise.

        quiet : bool, optional
            True if the state should be set without emitted the toggled
            signal. The default is False.

        """
        raise NotImplementedError

    def isEditable(self):
        """ Get whether the title is user editable.

        Returns
        -------
        result : bool
            True if the title is user editable, False otherwise.

        """
        raise NotImplementedError

    def setEditable(self, editable):
        """ Set whether or not the title is user editable.

        Parameters
        ----------
        editable : bool
            True if the title is user editable, False otherwise.

        """
        raise NotImplementedError

    def isForceHidden(self):
        """ Get whether or not the title bar is force hidden.

        Returns
        -------
        result : bool
            Whether or not the title bar is force hidden.

        """
        raise NotImplementedError

    def setForceHidden(self, hidden):
        """ Set the force hidden state of the title bar.

        Parameters
        ----------
        hidden : bool
            True if the title bar should be hidden, False otherwise.

        """
        raise NotImplementedError


class QDockTitleBar(QFrame, IDockTitleBar):
    """ A concrete implementation of IDockTitleBar.

    This class serves as the default title bar for a QDockItem.

    """
    #: A signal emitted when the maximize button is clicked.
    maximizeButtonClicked = Signal(bool)

    #: A signal emitted when the restore button is clicked.
    restoreButtonClicked = Signal(bool)

    #: A signal emitted when the close button is clicked.
    closeButtonClicked = Signal(bool)

    #: A signal emitted when the link button is toggled.
    linkButtonToggled = Signal(bool)

    #: A signal emitted when the pin button is toggled.
    pinButtonToggled = Signal(bool)

    #: A signal emitted when the tabs button is clicked.
    tabsButtonClicked = Signal(bool)

    #: A signal emitted when the title is edited by the user.
    titleEdited = Signal(str)

    #: A signal emitted when the empty area is left double clicked.
    leftDoubleClicked = Signal(QPoint)

    #: A signal emitted when the empty area is right clicked.
    rightClicked = Signal(QPoint)

    def __init__(self, parent=None):
        """ Initialize a QDockTitleBar.

        Parameters
        ----------
        parent : QWidget or None
            The parent of the title bar.

        """
        super(QDockTitleBar, self).__init__(parent)
        self._buttons = self.CloseButton | self.MaximizeButton | self.PinButton
        self._is_editable = False
        self._force_hidden = False
        self._last_visible = True
        self._line_edit = None

        title_icon = self._title_icon = QIconWidget(self)
        title_icon.setVisible(False)

        title_label = self._title_label = QTextLabel(self)

        spacer = self._spacer = QWidget(self)
        policy = spacer.sizePolicy()
        policy.setHorizontalPolicy(QSizePolicy.Expanding)
        spacer.setSizePolicy(policy)

        btn_size = QSize(14, 13)

        max_button = self._max_button = QBitmapButton(self)
        max_button.setObjectName('docktitlebar-maximize-button')
        max_button.setBitmap(MAXIMIZE_BUTTON.toBitmap())
        max_button.setIconSize(btn_size)
        max_button.setVisible(self._buttons & self.MaximizeButton)
        max_button.setToolTip('Maximize')

        restore_button = self._restore_button = QBitmapButton(self)
        restore_button.setObjectName('docktitlebar-restore-button')
        restore_button.setBitmap(RESTORE_BUTTON.toBitmap())
        restore_button.setIconSize(btn_size)
        restore_button.setVisible(self._buttons & self.RestoreButton)
        restore_button.setToolTip('Restore Down')

        close_button = self._close_button = QBitmapButton(self)
        close_button.setObjectName('docktitlebar-close-button')
        close_button.setBitmap(CLOSE_BUTTON.toBitmap())
        close_button.setIconSize(btn_size)
        close_button.setVisible(self._buttons & self.CloseButton)
        close_button.setToolTip('Close')

        link_button = self._link_button = QCheckedBitmapButton(self)
        link_button.setObjectName('docktitlebar-link-button')
        link_button.setBitmap(UNLINKED_BUTTON.toBitmap())
        link_button.setCheckedBitmap(LINKED_BUTTON.toBitmap())
        link_button.setIconSize(btn_size)
        link_button.setVisible(self._buttons & self.LinkButton)
        link_button.setToolTip('Link Window')
        link_button.setCheckedToolTip('Unlink Window')

        pin_button = self._pin_button = QCheckedBitmapButton(self)
        pin_button.setObjectName('docktitlebar-pin-button')
        pin_button.setBitmap(PIN_BUTTON.toBitmap())
        pin_button.setCheckedBitmap(UNPIN_BUTTON.toBitmap())
        pin_button.setIconSize(QSize(13, 13))
        pin_button.setVisible(self._buttons & self.PinButton)
        pin_button.setToolTip('Pin Window')
        pin_button.setCheckedToolTip('Unpin Window')

        tabs_button = self._tabs_button = QBitmapButton(self)
        tabs_button.setObjectName('docktitlebar-tabs-button')
        tabs_button.setBitmap(TAB_BUTTON.toBitmap())
        tabs_button.setIconSize(btn_size)
        tabs_button.setVisible(self._buttons & self.TabsButton)
        tabs_button.setToolTip('List all tabs')

        layout = QHBoxLayout()
        layout.setContentsMargins(QMargins(5, 2, 5, 2))
        layout.setSpacing(1)
        layout.addWidget(title_icon)
        layout.addSpacing(0)
        layout.addWidget(title_label)
        layout.addWidget(spacer)
        layout.addSpacing(4)
        layout.addWidget(tabs_button)
        layout.addWidget(pin_button)
        layout.addWidget(link_button)
        layout.addWidget(max_button)
        layout.addWidget(restore_button)
        layout.addWidget(close_button)

        self.setLayout(layout)

        max_button.clicked.connect(self.maximizeButtonClicked)
        restore_button.clicked.connect(self.restoreButtonClicked)
        close_button.clicked.connect(self.closeButtonClicked)
        link_button.toggled.connect(self.linkButtonToggled)
        pin_button.toggled.connect(self.pinButtonToggled)
        tabs_button.clicked.connect(self.tabsButtonClicked)

    #--------------------------------------------------------------------------
    # Event Handlers
    #--------------------------------------------------------------------------
    def mouseDoubleClickEvent(self, event):
        """ Handle the mouse double click event for the title bar.

        """
        event.ignore()
        if event.button() == Qt.LeftButton:
            pos = event.pos()
            is_editable = self._is_editable
            if self._adjustedLabelGeometry().contains(pos) and is_editable:
                self._showTitleLineEdit()
                event.accept()
                return
            if self._clickableGeometry().contains(pos):
                self.leftDoubleClicked.emit(event.globalPos())
                event.accept()
                return

    def mousePressEvent(self, event):
        """ Handle the mouse press event for the title bar.

        """
        event.ignore()
        if event.button() == Qt.RightButton:
            if self._clickableGeometry().contains(event.pos()):
                self.rightClicked.emit(event.globalPos())
                event.accept()
                return

    #--------------------------------------------------------------------------
    # Overrides
    #--------------------------------------------------------------------------
    def setVisible(self, visible):
        """ An overridden virtual visibility setter.

        This handler enforces the force-hidden setting.

        """
        self._last_visible = visible
        if visible and self._force_hidden:
            return
        super(QDockTitleBar, self).setVisible(visible)

    #--------------------------------------------------------------------------
    # Private API
    #--------------------------------------------------------------------------
    def _adjustedLabelGeometry(self):
        """ Get the adjust label geometry.

        Returns
        -------
        result : QRect
            A rectangle representing the label geometry which has been
            adjusted for potentially empty text. This rect can be used
            for a usable hit-testing rect for the label text.

        """
        label = self._title_label
        label_geo = label.geometry()
        if not label.text():
            label_geo = label_geo.adjusted(0, 0, 10, 0)
        return label_geo

    def _clickableGeometry(self):
        """ Get the geometry rect which represents clickable area.

        Returns
        -------
        result : QRect
            A rectangle adjusted for the clickable geometry.

        """
        rect = self.rect().adjusted(5, 2, -5, -2)
        rect.setRight(self._spacer.geometry().right())
        return rect

    def _showTitleLineEdit(self):
        """ Setup the line edit widget for editing the title.

        """
        old_line_edit = self._line_edit
        if old_line_edit is not None:
            old_line_edit.hide()
            old_line_edit.deleteLater()
        line_edit = self._line_edit = QLineEdit(self)
        line_edit.setFrame(False)
        line_edit.setText(self._title_label.text())
        line_edit.selectAll()
        h = self._title_label.height()
        line_edit.setMinimumHeight(h)
        line_edit.setMaximumHeight(h)
        line_edit.editingFinished.connect(self._onEditingFinished)
        layout = self.layout()
        idx = layout.indexOf(self._spacer)
        layout.insertWidget(idx, line_edit)
        self._spacer.hide()
        self._title_label.hide()
        line_edit.show()
        line_edit.setFocus(Qt.MouseFocusReason)

    def _onEditingFinished(self):
        """ Handle the 'editingFinished' signal for title line edit.

        """
        line_edit = self._line_edit
        if line_edit is not None:
            text = line_edit.text()
            line_edit.hide()
            line_edit.deleteLater()
            self._line_edit = None
            changed = self._title_label.text() != text
            if changed:
                self._title_label.setText(text)
            self._title_label.show()
            self._spacer.show()
            if changed:
                self.titleEdited.emit(text)

    #--------------------------------------------------------------------------
    # IDockItemTitleBar API
    #--------------------------------------------------------------------------
    def buttons(self):
        """ Get the buttons to show in the title bar.

        Returns
        -------
        result : int
            An or'd combination of the buttons to show.

        """
        return self._buttons

    def setButtons(self, buttons):
        """ Set the buttons to show in the title bar.

        Parameters
        ----------
        buttons : int
            An or'd combination of the buttons to show.

        """
        self._buttons = buttons
        self._max_button.setVisible(buttons & self.MaximizeButton)
        self._restore_button.setVisible(buttons & self.RestoreButton)
        self._close_button.setVisible(buttons & self.CloseButton)
        self._link_button.setVisible(buttons & self.LinkButton)
        self._pin_button.setVisible(buttons & self.PinButton)
        self._tabs_button.setVisible(buttons & self.TabsButton)

    def title(self):
        """ Get the title string of the title bar.

        Returns
        -------
        result : unicode
            The unicode title string for the title bar.

        """
        return self._title_label.text()

    def setTitle(self, title):
        """ Set the title string of the title bar.

        Parameters
        ----------
        title : unicode
            The unicode string to use for the title bar.

        """
        self._title_label.setText(title)

    def label(self):
        """ Get the label which holds the title string.

        Returns
        -------
        result : QTextLabel
            The label widget which holds the title string.

        """
        return self._title_label

    def icon(self):
        """ Get the icon for the title bar.

        Returns
        -------
        result : QIcon
            The icon set for the title bar.

        """
        return self._title_icon.icon()

    def setIcon(self, icon):
        """ Set the icon for the title bar.

        Parameters
        ----------
        icon : QIcon
            The icon to use for the title bar.

        """
        visible, spacing = (False, 0) if icon.isNull() else (True, 4)
        title_icon = self._title_icon
        title_icon.setIcon(icon)
        title_icon.setVisible(visible)
        layout = self.layout()
        layout.takeAt(1)
        layout.insertSpacing(1, spacing)

    def iconSize(self):
        """ Get the icon size for the title bar.

        Returns
        -------
        result : QSize
            The size to use for the icons in the title bar.

        """
        return self._title_icon.iconSize()

    def setIconSize(self, size):
        """ Set the icon size for the title bar.

        Parameters
        ----------
        icon : QSize
            The icon size to use for the title bar. Icons smaller than
            this size will not be scaled up.

        """
        self._title_icon.setIconSize(size)

    def isLinked(self):
        """ Get whether the link button is checked.

        Returns
        -------
        result : bool
            True if the link button is checked, False otherwise.

        """
        return self._link_button.isChecked()

    def setLinked(self, linked):
        """ Set whether or not the link button is checked.

        Parameters
        ----------
        linked : bool
            True if the link button should be checked, False otherwise.

        """
        self._link_button.setChecked(linked)

    def isPinned(self):
        """ Get whether the pin button is checked.

        Returns
        -------
        result : bool
            True if the pin button is checked, False otherwise.

        """
        return self._pin_button.isChecked()

    def setPinned(self, pinned, quiet=False):
        """ Set whether or not the pin button is checked.

        Parameters
        ----------
        pinned : bool
            True if the pin button should be checked, False otherwise.

        quiet : bool, optional
            True if the state should be set without emitted the toggled
            signal. The default is False.

        """
        old = self._pin_button.blockSignals(quiet)
        self._pin_button.setChecked(pinned)
        self._pin_button.blockSignals(old)

    def isEditable(self):
        """ Get whether the title is user editable.

        Returns
        -------
        result : bool
            True if the title is user editable, False otherwise.

        """
        return self._is_editable

    def setEditable(self, editable):
        """ Set whether or not the title is user editable.

        Parameters
        ----------
        editable : bool
            True if the title is user editable, False otherwise.

        """
        self._is_editable = editable

    def isForceHidden(self):
        """ Get whether or not the title bar is force hidden.

        Returns
        -------
        result : bool
            Whether or not the title bar is always hidden.

        """
        return self._force_hidden

    def setForceHidden(self, hidden):
        """ Set the force hidden state of the title bar.

        Parameters
        ----------
        hidden : bool
            True if the title bar should be hidden, False otherwise.

        """
        self._force_hidden = hidden
        if not hidden and self._last_visible:
            super(QDockTitleBar, self).setVisible(True)
        elif hidden:
            super(QDockTitleBar, self).setVisible(False)



#------------------------------------------------------------------------------
# Copyright (c) 2013, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
import re

from atom.api import Atom, Bool, Typed, Enum, Str, Str


class Validator(Atom):
    """ The base class for creating widget text validators.

    This class is abstract. It's abstract api must be implemented by a
    subclass in order to be usable.

    """
    #: An optional message to associate with the validator. This message
    #: may be used by the toolkit to display information to the user
    #: about what went wrong.
    message = Str()

    def validate(self, text):
        """ Validate the text as the user types.

        This method is called on every keystroke to validate the text
        as the user inputs characters. It should be efficient. This is
        an abstract method which must be implemented by subclasses.

        Parameters
        ----------
        text : unicode
            The unicode text entered by the user.

        Returns
        -------
        result : bool
            True if the text is valid, False otherwise.

        """
        raise NotImplementedError

    def fixup(self, text):
        """ An optional method to fix invalid user input.

        This method will be called if user attempts to apply text which
        is not valid. This method may convert the text to a valid form.
        The returned text will be retested for validity. The default
        implementation of this method is a no-op.

        Returns
        -------
        result : unicode
            The optionally modified input text.

        """
        return text


class IntValidator(Validator):
    """ A concrete Validator which handles integer input.

    This validator ensures that the text represents an integer within a
    specified range in a specified base.

    """
    #: The minimum value allowed for the int, inclusive, or None if
    #: there is no lower bound.
    minimum = Typed(int)

    #: The maximum value allowed for the int, inclusive, or None if
    #: there is no upper bound.
    maximum = Typed(int)

    #: The base in which the int is represented.
    base = Enum(10, 2, 8, 16)

    def validate(self, text):
        """ Validates the given text matches the integer range.

        Parameters
        ----------
        text : unicode
            The unicode text edited by the client widget.

        Returns
        -------
        result : bool
            True if the text is valid, False otherwise.

        """
        try:
            value = int(text, self.base)
        except ValueError:
            return False
        minimum = self.minimum
        if minimum is not None and value < minimum:
            return False
        maximum = self.maximum
        if maximum is not None and value > maximum:
            return False
        return True


class FloatValidator(Validator):
    """ A concrete Validator which handles floating point input.

    This validator ensures that the text represents a floating point
    number within a specified range.

    """
    #: The minimum value allowed for the float, inclusive, or None if
    #: there is no lower bound.
    minimum = Typed(float)

    #: The maximum value allowed for the float, inclusive, or None if
    #: there is no upper bound.
    maximum = Typed(float)

    #: Whether or not to allow exponents like '1e6' in the input.
    allow_exponent = Bool(True)

    def validate(self, text):
        """ Validates the given text matches the float range.

        Parameters
        ----------
        text : unicode
            The unicode text edited by the client widget.

        Returns
        -------
        result : bool
            True if the text is valid, False otherwise.

        """
        try:
            value = float(text)
        except ValueError:
            return False
        minimum = self.minimum
        if minimum is not None and value < minimum:
            return False
        maximum = self.maximum
        if maximum is not None and value > maximum:
            return False
        if not self.allow_exponent and 'e' in text.lower():
            return False
        return True


class RegexValidator(Validator):
    """ A concrete Validator which handles text input.

    This validator ensures that the text matches a provided regular
    expression string.

    """
    #: The regular expression string to use for validation. The default
    #: regex matches everything.
    regex = Str(r'.*')

    def validate(self, text):
        """ Validates the given text matches the regular expression.

        Parameters
        ----------
        text : unicode
            The unicode text edited by the client widget.

        Returns
        -------
        result : bool
            True if the text is valid, False otherwise.

        """
        return bool(re.match(self.regex, text))



#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------


def repolish(widget):
    """ Repolish a widget when the stylesheet dependencies change.

    Parameters
    ----------
    widget : QWidget
        The widget to repolish.

    """
    style = widget.style()
    style.unpolish(widget)
    style.polish(widget)
